/*
 * Program.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

//
// NOTE: You will need this using statement for various things, including the
//       Console class.
//
using System;
using System.Threading;

//
// NOTE: Typically, you will need all of the following using statements to take
//       full advantage of the Eagle library.
//
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

///////////////////////////////////////////////////////////////////////////////

namespace TclDotNet.Tutorial
{
    //
    // NOTE: Code goes here.
    //
    ///////////////////////////////////////////////////////////////////////////

    #region Custom Host Class
    internal class CustomHost : Eagle._Hosts.Console
    {
        #region Public Constructors
        public CustomHost(
            IHostData hostData
            )
            : base(hostData)
        {
            // do nothing.
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IHost Members
        public override bool CanExit
        {
            get { return false; }
        }
        #endregion
    }
    #endregion

    ///////////////////////////////////////////////////////////////////////////

    #region Custom Command Class
    [ObjectId("0b943624-25cd-4df1-bf99-1044a3bfcad8")]
    [CommandFlags(CommandFlags.None)]
    [ObjectGroup("demo")]
    internal class CustomCommand : IExecute
    {
        #region Private Data
        private long id;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IExecute Members
        public ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            if (id == 0)
                id = interpreter.NextId();

            result = StringList.MakeList(id, DateTime.UtcNow);
            return ReturnCode.Ok;
        }
        #endregion
    }
    #endregion

    ///////////////////////////////////////////////////////////////////////////

    #region Custom Resolver Class
    internal class CustomResolver : IResolve
    {
        #region Private Data
        private IResolve resolve;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public CustomResolver(
            IResolveData resolveData,
            IResolve resolve
            )
        {
            if (resolveData != null)
                interpreter = resolveData.Interpreter;

            this.resolve = resolve;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IResolveData Members
        private ResolveFlags flags;
        public ResolveFlags Flags
        {
            get { return flags; }
            set { flags = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IResolve Members
        public ReturnCode GetIExecute(
            ICallFrame frame,
            EngineFlags engineFlags,
            string name,
            ArgumentList arguments,
            LookupFlags lookupFlags,
            ref bool ambiguous,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (name == "not_found")
            {
                error = "command not found";
                return ReturnCode.Return;
            }

            if (name == "bad_execute")
            {
                execute = null;
                return ReturnCode.Ok;
            }

            if (name == "magic_execute")
            {
                execute = new CustomCommand();
                return ReturnCode.Ok;
            }

            if (resolve == null)
            {
                error = "invalid fallback resolver";
                return ReturnCode.Error;
            }

            return resolve.GetIExecute(
                frame, engineFlags, name, arguments, lookupFlags,
                ref ambiguous, ref token, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public ReturnCode GetVariable(
            ICallFrame frame,
            string varName,
            string varIndex,
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            )
        {
            IResolve resolve = null;

            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (resolve == null)
            {
                error = "invalid fallback resolver";
                return ReturnCode.Error;
            }

            return resolve.GetVariable(
                frame, varName, varIndex, ref flags,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public ReturnCode GetVariableFrame(
            ref ICallFrame frame,
            ref string varName,
            ref VariableFlags flags,
            ref Result error
            )
        {
            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        public ReturnCode GetCurrentNamespace(
            ICallFrame frame,
            ref INamespace @namespace,
            ref Result error
            )
        {
            error = "not implemented";
            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IIdentifier Members
        public IClientData ClientData
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        ///////////////////////////////////////////////////////////////////////

        public string Description
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        ///////////////////////////////////////////////////////////////////////

        public string Group
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        ///////////////////////////////////////////////////////////////////////

        public Guid Id
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        ///////////////////////////////////////////////////////////////////////

        public IdentifierKind Kind
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }

        ///////////////////////////////////////////////////////////////////////

        public string Name
        {
            get { throw new NotImplementedException(); }
            set { throw new NotImplementedException(); }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IGetInterpreter / ISetInterpreter Members
        private Interpreter interpreter;
        public Interpreter Interpreter
        {
            get { return interpreter; }
            set { interpreter = value; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IWrapperData Members
        private long token;
        public long Token
        {
            get { return token; }
            set { token = value; }
        }
        #endregion
    }
    #endregion

    ///////////////////////////////////////////////////////////////////////////

    #region Main Program Class
    internal static class Program
    {
        #region Entry Point Method
        private static void Main(
            string[] args
            )
        {
            ReturnCode code;
            Result result = null;

            using (Interpreter interpreter = Interpreter.Create(args, ref result))
            {
                if (interpreter != null)
                {
                    #region Setting Up Custom Host (Optional)
                    IHost host = new CustomHost(new HostData(
                        null, null, null, ClientData.Empty,
                        typeof(CustomHost).Name, interpreter,
                        null, null, HostCreateFlags.Default));

                    code = host.IsOpen() ?
                        ReturnCode.Ok : host.Open(ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;

                    interpreter.Host = host;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Show Some Windows Form (Optional)
                    Thread thread = null;

                    code = TclDotNetForm.Show(
                        interpreter, false, true, args, ref thread, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Setting Up Tcl
                    #region Phase 1: Load Tcl Library
                    code = interpreter.LoadTcl(
                        FindFlags.All, LoadFlags.None, null, null, null, null,
                        null, null, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Phase 2: Create Tcl Interpreter
                    string interpName = null;

                    code = interpreter.CreateTclInterpreter(
                        true, false, false, ref result);

                    WriteResult(code, result);

                    if (code == ReturnCode.Ok)
                        interpName = result;
                    else
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Phase 3: Create Bridged Tcl Command
                    code = interpreter.AddTclBridge(
                        new CustomCommand(), interpName, "foo", null, true,
                        true, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region The One-Liner REPL (Read-Eval-Print-Loop)
                    //
                    // NOTE: Here it is, the one-liner REPL.
                    //
                    code = Interpreter.InteractiveLoop(
                        interpreter, args, ref result);

                    ///////////////////////////////////////////////////////////

                    //
                    // NOTE: Optionally, we can display the overall result of
                    //       the interactive loop.
                    //
                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;

                    ///////////////////////////////////////////////////////////

                    //
                    // NOTE: Optionally, we can bring the Eagle interpreter
                    //       back to life.
                    //
                    interpreter.Exit = false;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Tearing Down Tcl
                    #region Phase 4: Delete Bridged Tcl Command
                    code = interpreter.RemoveTclBridge(
                        interpName, "foo", null, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Phase 5: Delete Tcl Interpreter
                    code = interpreter.DeleteTclInterpreter(
                        interpName, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Phase 6: Unload Tcl Library
                    code = interpreter.UnloadTcl(
                        UnloadFlags.Default, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Custom Resolver Setup
                    #region Prevent Unknown Handling
                    //
                    // NOTE: Optionally, we can prevent unknown commands from
                    //       triggering "unknown" handling.
                    //
                    interpreter.Unknown = null;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Create & Add Custom Resolver
                    //
                    // NOTE: Grab the default resolver now so that it can be
                    //       used as a fallback for our custom resolver.
                    //
                    IResolve resolve = null;

                    code = interpreter.GetResolver(
                        Priority.None, ref resolve, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;

                    code = interpreter.AddResolver(
                        new CustomResolver(new ResolveData(
                            null, null, null, null, interpreter,
                            ResolveFlags.None, 0), resolve),
                            null, Priority.None, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Demo Custom Resolver (REPL)
                    code = Interpreter.InteractiveLoop(
                        interpreter, args, ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;

                    interpreter.Exit = false;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Reset [To The Default] Resolver
                    code = interpreter.ResetResolvers(ref result);

                    WriteResult(code, result);

                    if (code != ReturnCode.Ok)
                        goto done;
                    #endregion
                    #endregion
                }
                else
                {
                    #region Error Handling (could not create Eagle interpreter?)
                    Console.WriteLine(Utility.FormatResult(
                        ReturnCode.Error, result));
                    #endregion
                }
            }

        done:
            Console.Write("Please press any key to continue.");
            Console.ReadKey();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Result Output Helper Method
        private static void WriteResult(
            ReturnCode code,
            Result result
            )
        {
            Console.WriteLine(Utility.FormatResult(code, result));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Tcl Script Evaluation Helper Method
        private static void EvaluateTclOneLiner(
            Interpreter interpreter,
            string interpName,
            ref Result result
            )
        {
            Console.Write("Please enter Tcl script to evaluate: ");

            WriteResult(interpreter.EvaluateTclScript(
                interpName, Console.ReadLine(), ref result), result);

            Console.Write("Please press any key to continue.");
            Console.ReadKey(true);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Eagle Script Evaluation Helper Method
        private static void EvaluateEagleOneLiner(
            Interpreter interpreter,
            ref Result result
            )
        {
            Console.Write("Please enter Eagle script to evaluate: ");

            WriteResult(interpreter.EvaluateScript(
                Console.ReadLine(), ref result), result);

            Console.Write("Please press any key to continue.");
            Console.ReadKey(true);
        }
        #endregion
    }
    #endregion
}
